from functools import wraps

from marshmallow import ValidationError
from werkzeug.exceptions import BadRequest

from fluidly.flask.api_exception import APIException
from fluidly.flask.rest_logger import rest_log_entrypoint


def handle_exceptions(func):
    @wraps(func)
    def decorated_function(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except APIException as exception:
            raise exception
        except ValidationError as exception:
            raise APIException(
                title="Schema Validation Error",
                status=422,
                detail=exception.normalized_messages(),
            )
        except BadRequest as exception:
            raise APIException(
                title="Invalid Body", status=400, detail=exception.description
            )
        except Exception:
            raise APIException(status=500, title="An unknown error occurred")
        return result

    return decorated_function


def log_safely(func):
    @rest_log_entrypoint
    @handle_exceptions
    @wraps(func)
    def decorated_function(*args, **kwargs):
        return func(*args, **kwargs)

    return decorated_function



import time
from functools import wraps

from flask import request

from fluidly.flask.api_exception import APIException
from fluidly.structlog import base_logger


def rest_log_entrypoint(func):
    @wraps(func)
    def decorated_function(*args, **kwargs):
        headers = dict(request.headers)

        connection_id_from_path = request.view_args.get("connection_id")
        partner_id_from_path = request.view_args.get("partner_id")

        logger = base_logger.get_logger()
        logger = logger.new(callback=func.__qualname__)
        start = time.time()

        try:
            result = func(*args, **kwargs)
        except APIException as exception:
            end = time.time()
            logger.error(
                "rest_request_processed",
                duration=end - start,
                success=False,
                exc_info=True,
                connection_id=connection_id_from_path,
                partner_id=partner_id_from_path,
                headers=headers,
                args=dict(request.view_args),
                status_code=exception.status,
                url=request.full_path,
            )
            raise exception

        end = time.time()
        logger.info(
            "rest_request_processed",
            duration=end - start,
            success=True,
            connection_id=connection_id_from_path,
            partner_id=partner_id_from_path,
            headers=headers,
            args=dict(request.view_args),
            status_code=result.status_code,
            url=request.full_path,
        )
        return result

    return decorated_function



from typing import Optional

from flask import Response, request

from fluidly.structlog.base_logger import get_logger

WILDCARD_ACCEPT_HEADER = "*/*"


def extract_root_type(content_type: Optional[str]) -> Optional[str]:
    """Examples of subset content types are:

    application/problem+json
    image/svg+xml
    """
    if not content_type:
        return content_type

    try:
        prefix, suffix = content_type.split("/", 1)
        _, base_suffix = suffix.rsplit("+", 1)
    except Exception:
        # If any exception occurs, do not prevent the response from returning
        return content_type

    return f"{prefix}/{base_suffix}"


def is_valid_content_type(
    accept_header: str, response_content_type: Optional[str]
) -> bool:
    return (
        accept_header == WILDCARD_ACCEPT_HEADER
        or accept_header == response_content_type
        or accept_header == extract_root_type(response_content_type)
    )


def validate_content_type(response: Response) -> Response:
    """Checks if the response content-type is compatible with the request's
    Accept header. If a mismatch is found, log the request.

    Use with flask.after_request hook i.e.

        app.after_request(validate_content_type)

    Note: this currently does very basic validation to cover the common cases.
    It does not support multiple mimetypes or wildcard subtypes (for now).

    If an exception is raised during the request, after_request hooks are not run.
    """
    headers = request.headers
    accept_header = headers.get("Accept", WILDCARD_ACCEPT_HEADER)

    accept_headers = accept_header.split(",")

    if not any(is_valid_content_type(h, response.content_type) for h in accept_headers):
        logger = get_logger()

        logger.warning(
            "Incompatible request Accept header and response content-type",
            args=dict(request.view_args) if request.view_args is not None else None,
            headers=dict(headers),
            url=request.full_path,
            request_accept_header=accept_header,
            response_content_type=response.content_type,
        )

    return response



import json
from typing import Dict, List, Optional, Union

from flask import Response


class APIException(Exception):
    def __init__(
        self, status: int, title: str, detail: Optional[Dict[str, List[str]]] = None
    ):
        """While using APIException don't forget to register custom error
        handler:

            app.register_error_handler(APIException, handle_api_exception)
        """
        self.title = title
        self.status = status
        self.detail = detail

    def to_dict(self) -> Dict[str, Optional[Union[str, int, Dict[str, List[str]]]]]:
        return {"title": self.title, "status": self.status, "detail": self.detail}


def handle_api_exception(error: APIException) -> Response:
    return Response(
        response=json.dumps(error.to_dict()),
        status=error.status,
        mimetype="application/problem+json",
    )



import base64


def base64_decode(encoded_str):
    """To make the string decodable you have to make the number of characters in the
    encoded string an integer multiple of 4. Meaning you have to divide the number of
    characters by 4 and not get a remainder.

    https://en.wikipedia.org/wiki/Base64#Output_padding
    """
    if isinstance(encoded_str, str):
        encoded_str = encoded_str.encode("utf-8")

    num_missed_paddings = 4 - len(encoded_str) % 4
    if num_missed_paddings != 4:
        encoded_str += b"=" * num_missed_paddings
    return base64.b64decode(encoded_str).decode("utf-8")



import json
from functools import wraps

from flask import g, request

from fluidly.auth.permissions import (
    UserPermissionsPayloadException,
    UserPermissionsRequestException,
    check_admin_permissions,
    check_user_permissions,
)
from fluidly.flask.api_exception import APIException
from fluidly.flask.utils import base64_decode


def authorised(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Retrieves the authentication information from Google Cloud Endpoints
        and passes it to user permissions service"""
        encoded_user_info = request.headers.get("X-Endpoint-API-UserInfo", None)
        if not encoded_user_info:
            raise APIException(status=401, title="User is not authenticated")

        decoded_user_info = base64_decode(encoded_user_info)
        user_info = json.loads(decoded_user_info)

        if "claims" in user_info:
            claims = json.loads(user_info.get("claims", "{}"))
        else:
            # Support ESPv2
            claims = user_info

        auth0_claims = claims.get("https://api.fluidly.com/app_metadata", {})
        internal_claims = claims.get("https://api.fluidly.com/internal_metadata", {})

        connection_id = request.view_args["connection_id"]
        user_id = auth0_claims.get("userId", None)

        try:
            is_service_account = internal_claims.get("isServiceAccount", False)

            if not is_service_account and not check_user_permissions(
                claims, connection_id
            ):
                raise APIException(status=403, title="User cannot access this resource")
        except (
            ValueError,
            UserPermissionsPayloadException,
            UserPermissionsRequestException,
        ):
            raise APIException(
                status=403, title="An issue occurred while fetching permissions"
            )

        g.connection_id = connection_id
        g.user_id = user_id
        return f(*args, **kwargs)

    return decorated_function


def admin(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Retrieves the authentication information from Google Cloud Endpoints and passes it to user permissions service"""
        encoded_info = request.headers.get("X-Endpoint-API-UserInfo", None)
        if not encoded_info:
            raise APIException(status=401, title="User is not authenticated")

        info_json = base64_decode(encoded_info)
        # First parsing of the decoded header string
        user_info = json.loads(info_json)
        # Claims are given as a string by Cloud Endpoints so we have
        # to parse the claims attribute
        if "claims" in user_info:
            claims = json.loads(user_info.get("claims", "{}"))
        else:
            # Support ESPv2
            claims = user_info

        auth0_claims = claims.get("https://api.fluidly.com/app_metadata", {})
        internal_claims = claims.get("https://api.fluidly.com/internal_metadata", {})

        user_id_from_token = auth0_claims.get("userId", None)

        try:
            is_service_account = internal_claims.get("isServiceAccount", False)

            if not is_service_account and not check_admin_permissions(claims):
                raise APIException(status=403, title="User cannot access this resource")
        except (
            ValueError,
            UserPermissionsPayloadException,
            UserPermissionsRequestException,
        ):
            raise APIException(
                status=403, title="An issue occurred while fetching permissions"
            )

        g.user_id = user_id_from_token
        return f(*args, **kwargs)

    return decorated_function
